@@ -268,7 +268,7 @@ minSetShade (CompWindow *w,
     XIntersectRegion (mw->region, &rect, w->region);
     XOffsetRegion (w->region, w->attrib.x, w->attrib.y - (h - shade));
 
-    w->matrix = w->texture.matrix;
+    w->matrix = w->texture->matrix;
     w->matrix.x0 -= (w->attrib.x * w->matrix.xx);
     w->matrix.y0 -= ((w->attrib.y - (h - shade)) * w->matrix.yy);
 
@@ -671,7 +671,7 @@ minHandleEvent (CompDisplay *d,
 	{
 	    MIN_SCREEN (w->screen);
 
-	    if (w->pendingUnmaps) /* Normal -> Iconic */
+	    if (w->pendingUnmaps && onCurrentDesktop (w)) /* Normal -> Iconic */
 	    {
 		MIN_WINDOW (w);
 
@@ -828,7 +828,7 @@ minDamageWindowRect (CompWindow *w,
 		XOffsetRegion (mw->region, -w->attrib.x, -w->attrib.y);
 
 		/* bind pixmap here so we have something to unshade with */
-		if (!w->texture.pixmap)
+		if (!w->texture->pixmap)
 		    bindWindow (w);
 
 		ms->moreAdjust = TRUE;
@@ -985,17 +985,26 @@ minInitWindow (CompPlugin *p,
 
     mw->unmapCnt = 0;
 
-    if (w->shaded)
+    if (w->state & CompWindowStateHiddenMask)
     {
-	mw->state = mw->newState = NormalState;
-	mw->shade = 0;
-	mw->region = XCreateRegion ();
+	if (w->shaded)
+	{
+	    mw->state = mw->newState = NormalState;
+	    mw->shade = 0;
+	    mw->region = XCreateRegion ();
+	}
+	else
+	{
+	    mw->state = mw->newState = minGetWindowState (w);
+	    mw->shade = MAXSHORT;
+	    mw->region = NULL;
+	}
     }
     else
     {
-	mw->state = mw->newState = minGetWindowState (w);
-	mw->shade = MAXSHORT;
-	mw->region = NULL;
+	    mw->state = mw->newState = NormalState;
+	    mw->shade = MAXSHORT;
+	    mw->region = NULL;
     }
 
     w->privates[ms->windowPrivateIndex].ptr = mw;
@@ -1065,7 +1074,9 @@ static CompPluginVTable minVTable = {
     minGetScreenOptions,
     minSetScreenOption,
     minDeps,
-    sizeof (minDeps) / sizeof (minDeps[0])
+    sizeof (minDeps) / sizeof (minDeps[0]),
+    0, /* Features */
+    0  /* nFeatures */
 };
 
 CompPluginVTable *
