@@ -77,6 +77,9 @@ static CompWatchFdHandle lastWatchFdHandle = 1;
 static struct pollfd     *watchPollFds = 0;
 static int               nWatchFds = 0;
 
+static CompScreen *targetScreen = NULL;
+static int        targetOutput = 0;
+
 static Bool inHandleEvent = FALSE;
 
 static Bool shutDown = FALSE;
@@ -2005,6 +2008,8 @@ eventLoop (void)
 		    if (!s->damageMask || s->timeLeft > timeToNextRedraw)
 			continue;
 
+		    targetScreen = s;
+
 		    timeDiff = TIMEVALDIFF (&tv, &s->lastRedraw);
 
 		    /* handle clock rollback */
@@ -2065,15 +2070,23 @@ eventLoop (void)
 		    mask = s->damageMask;
 		    s->damageMask = 0;
 
+		    if (s->clearBuffers)
+		    {
+			if (mask & COMP_SCREEN_DAMAGE_ALL_MASK)
+			    glClear (GL_COLOR_BUFFER_BIT);
+		    }
+
 		    for (i = 0; i < s->nOutputDev; i++)
 		    {
+			targetScreen = s;
+			targetOutput = i;
+
 			if (s->nOutputDev > 1)
-			    glScissor (s->outputDev[i].region.extents.x1,
-				       s->outputDev[i].region.extents.y1,
-				       s->outputDev[i].region.extents.x2 -
-				       s->outputDev[i].region.extents.x1,
-				       s->outputDev[i].region.extents.y2 -
-				       s->outputDev[i].region.extents.y1);
+			    glViewport (s->outputDev[i].region.extents.x1,
+					s->height -
+					s->outputDev[i].region.extents.y2,
+					s->outputDev[i].width,
+					s->outputDev[i].height);
 
 			if (mask & COMP_SCREEN_DAMAGE_ALL_MASK)
 			{
@@ -2082,12 +2095,6 @@ eventLoop (void)
 					       &s->outputDev[i].region, i,
 					       PAINT_SCREEN_REGION_MASK |
 					       PAINT_SCREEN_FULL_MASK);
-
-			    if (i + 1 == s->nOutputDev)
-			    {
-				waitForVideoSync (s);
-				glXSwapBuffers (s->display->display, s->output);
-			    }
 			}
 			else if (mask & COMP_SCREEN_DAMAGE_REGION_MASK)
 			{
@@ -2095,87 +2102,93 @@ eventLoop (void)
 					      &s->outputDev[i].region,
 					      outputRegion);
 
-			    if ((*s->paintScreen) (s,
-						   &defaultScreenPaintAttrib,
-						   outputRegion, i,
-						   PAINT_SCREEN_REGION_MASK))
-			    {
-				BoxPtr pBox;
-				int    nBox, y;
-
-				pBox = outputRegion->rects;
-				nBox = outputRegion->numRects;
-
-				waitForVideoSync (s);
-
-				if (s->copySubBuffer)
-				{
-				    while (nBox--)
-				    {
-					y = s->height - pBox->y2;
-
-					(*s->copySubBuffer) (display->display,
-							     s->output,
-							     pBox->x1, y,
-							     pBox->x2 -
-							     pBox->x1,
-							     pBox->y2 -
-							     pBox->y1);
-
-					pBox++;
-				    }
-				}
-				else
-				{
-				    glEnable (GL_SCISSOR_TEST);
-				    glDrawBuffer (GL_FRONT);
-
-				    while (nBox--)
-				    {
-					y = s->height - pBox->y2;
-
-					glBitmap (0, 0, 0, 0,
-						  pBox->x1 - s->rasterX,
-						  y - s->rasterY,
-						  NULL);
-
-					s->rasterX = pBox->x1;
-					s->rasterY = y;
-
-					glScissor (pBox->x1, y,
-						   pBox->x2 - pBox->x1,
-						   pBox->y2 - pBox->y1);
-
-					glCopyPixels (pBox->x1, y,
-						      pBox->x2 - pBox->x1,
-						      pBox->y2 - pBox->y1,
-						      GL_COLOR);
-
-					pBox++;
-				    }
-
-				    glDrawBuffer (GL_BACK);
-				    glDisable (GL_SCISSOR_TEST);
-				    glFlush ();
-				}
-			    }
-			    else
+			    if (!(*s->paintScreen) (s,
+						    &defaultScreenPaintAttrib,
+						    outputRegion, i,
+						    PAINT_SCREEN_REGION_MASK))
 			    {
 				(*s->paintScreen) (s,
 						   &defaultScreenPaintAttrib,
 						   &s->outputDev[i].region, i,
 						   PAINT_SCREEN_FULL_MASK);
 
-				if (i + 1 == s->nOutputDev)
-				{
-				    waitForVideoSync (s);
-				    glXSwapBuffers (display->display,
-						    s->output);
-				}
+				XUnionRegion (tmpRegion,
+					      &s->outputDev[i].region,
+					      tmpRegion);
+
 			    }
 			}
 		    }
 
+		    targetScreen = NULL;
+		    targetOutput = 0;
+
+		    waitForVideoSync (s);
+
+		    if (mask & COMP_SCREEN_DAMAGE_ALL_MASK)
+		    {
+			glXSwapBuffers (display->display, s->output);
+		    }
+		    else
+		    {
+			BoxPtr pBox;
+			int    nBox, y;
+
+			pBox = tmpRegion->rects;
+			nBox = tmpRegion->numRects;
+
+			if (s->copySubBuffer)
+			{
+			    while (nBox--)
+			    {
+				y = s->height - pBox->y2;
+
+				(*s->copySubBuffer) (display->display,
+						     s->output,
+						     pBox->x1, y,
+						     pBox->x2 -
+						     pBox->x1,
+						     pBox->y2 -
+						     pBox->y1);
+
+				pBox++;
+			    }
+			}
+			else
+			{
+			    glEnable (GL_SCISSOR_TEST);
+			    glDrawBuffer (GL_FRONT);
+
+			    while (nBox--)
+			    {
+				y = s->height - pBox->y2;
+
+				glBitmap (0, 0, 0, 0,
+					  pBox->x1 - s->rasterX,
+					  y - s->rasterY,
+					  NULL);
+
+				s->rasterX = pBox->x1;
+				s->rasterY = y;
+
+				glScissor (pBox->x1, y,
+					   pBox->x2 - pBox->x1,
+					   pBox->y2 - pBox->y1);
+
+				glCopyPixels (pBox->x1, y,
+					      pBox->x2 - pBox->x1,
+					      pBox->y2 - pBox->y1,
+					      GL_COLOR);
+
+				pBox++;
+			    }
+
+			    glDrawBuffer (GL_BACK);
+			    glDisable (GL_SCISSOR_TEST);
+			    glFlush ();
+			}
+		    }
+
 		    s->lastRedraw = tv;
 
 		    (*s->donePaintScreen) (s);
@@ -2506,7 +2519,8 @@ addDisplay (char *name,
     d->initPluginForDisplay = initPluginForDisplay;
     d->finiPluginForDisplay = finiPluginForDisplay;
 
-    d->handleEvent = handleEvent;
+    d->handleEvent	 = handleEvent;
+    d->handleCompizEvent = handleCompizEvent;
 
     d->supportedAtom	     = XInternAtom (dpy, "_NET_SUPPORTED", 0);
     d->supportingWmCheckAtom = XInternAtom (dpy, "_NET_SUPPORTING_WM_CHECK", 0);
@@ -2584,20 +2598,25 @@ addDisplay (char *name,
     d->winStateDisplayModalAtom	    =
 	XInternAtom (dpy, "_NET_WM_STATE_DISPLAY_MODAL", 0);
 
-    d->winActionMoveAtom	 = XInternAtom (dpy, "_NET_WM_ACTION_MOVE", 0);
-    d->winActionResizeAtom	 =
+    d->winActionMoveAtom	  = XInternAtom (dpy, "_NET_WM_ACTION_MOVE", 0);
+    d->winActionResizeAtom	  =
 	XInternAtom (dpy, "_NET_WM_ACTION_RESIZE", 0);
-    d->winActionStickAtom	 = XInternAtom (dpy, "_NET_WM_ACTION_STICK", 0);
-    d->winActionMinimizeAtom	 =
+    d->winActionStickAtom	  =
+	XInternAtom (dpy, "_NET_WM_ACTION_STICK", 0);
+    d->winActionMinimizeAtom	  =
 	XInternAtom (dpy, "_NET_WM_ACTION_MINIMIZE", 0);
-    d->winActionMaximizeHorzAtom =
+    d->winActionMaximizeHorzAtom  =
 	XInternAtom (dpy, "_NET_WM_ACTION_MAXIMIZE_HORZ", 0);
-    d->winActionMaximizeVertAtom =
+    d->winActionMaximizeVertAtom  =
 	XInternAtom (dpy, "_NET_WM_ACTION_MAXIMIZE_VERT", 0);
-    d->winActionFullscreenAtom	 =
+    d->winActionFullscreenAtom	  =
 	XInternAtom (dpy, "_NET_WM_ACTION_FULLSCREEN", 0);
-    d->winActionCloseAtom	 = XInternAtom (dpy, "_NET_WM_ACTION_CLOSE", 0);
-    d->winActionShadeAtom	 = XInternAtom (dpy, "_NET_WM_ACTION_SHADE", 0);
+    d->winActionCloseAtom	  =
+	XInternAtom (dpy, "_NET_WM_ACTION_CLOSE", 0);
+    d->winActionShadeAtom	  =
+	XInternAtom (dpy, "_NET_WM_ACTION_SHADE", 0);
+    d->winActionChangeDesktopAtom =
+	XInternAtom (dpy, "_NET_WM_ACTION_CHANGE_DESKTOP", 0);
 
     d->wmAllowedActionsAtom = XInternAtom (dpy, "_NET_WM_ALLOWED_ACTIONS", 0);
 
@@ -3325,3 +3344,13 @@ setDisplayAction (CompDisplay     *display,
 
     return FALSE;
 }
+
+void
+clearTargetOutput (CompDisplay	*display,
+		   unsigned int mask)
+{
+    if (targetScreen)
+	clearScreenOutput (targetScreen,
+			   targetOutput,
+			   mask);
+}
