@@ -80,6 +80,7 @@ CompWindow *lastDamagedWindow = 0;
 Bool replaceCurrentWm = FALSE;
 Bool indirectRendering = FALSE;
 Bool strictBinding = FALSE;
+Bool noDetection = FALSE;
 
 #ifdef USE_COW
 Bool useCow = FALSE;
@@ -95,10 +96,10 @@ usage (void)
 	    "[--fast-filter] "
 	    "[--indirect-rendering] "
 	    "[--strict-binding] "
-	    "[--test-mode]\n       "
-	    "[--replace] "
+	    "[--replace]\n       "
 	    "[--sm-disable] "
 	    "[--sm-client-id ID] "
+	    "[--no-detection] "
 	    "[--version]\n       "
 
 #ifdef USE_COW
@@ -216,6 +217,10 @@ main (int argc, char **argv)
 	    if (i + 1 < argc)
 		clientId = argv[++i];
 	}
+	else if (!strcmp (argv[i], "--no-detection"))
+	{
+	    noDetection = TRUE;
+	}
 	else if (!strcmp (argv[i], "--bg-image"))
 	{
 	    if (i + 1 < argc)
