@@ -214,6 +214,12 @@ compSetActionOption (CompOption      *option,
 		equal = FALSE;
 	}
 
+	if (value->action.type & CompBindingTypeEdgeButton)
+	{
+	    if (action->edgeButton != value->action.edgeButton)
+		equal = FALSE;
+	}
+
 	if (equal)
 	    return FALSE;
     }
@@ -421,6 +427,24 @@ getStringOptionNamed (CompOption *option,
     return defaultValue;
 }
 
+unsigned short *
+getColorOptionNamed (CompOption	    *option,
+		     int	    nOption,
+		     char	    *name,
+		     unsigned short *defaultValue)
+{
+    while (nOption--)
+    {
+	if (option->type == CompOptionTypeColor)
+	    if (strcmp (option->name, name) == 0)
+		return option->value.c;
+
+	option++;
+    }
+
+    return defaultValue;
+}
+
 static char *
 stringAppend (char *s,
 	      char *a)
